/*
 *  
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/videodev.h>

#include "pvrusb2.h"
#include "pvrusb2-v4l2.h"
#include "pvrusb2-saa7115.h"
#include "pvrusb2-msp34x5g.h"
#include "pvrusb2-tuner.h"
#include "pvrusb2-buffer.h"
#include "pvrusb2-stream.h"

struct v4l2_capability pvr_capability ={
    .driver         = "pvrusb2",
    .card           = "Hauppauge WinTV pvr-usb2",
    .bus_info       = "usb",
    .version        = KERNEL_VERSION(0,6,0),
    .capabilities   = V4L2_CAP_VIDEO_CAPTURE | V4L2_CAP_VBI_CAPTURE
                      | V4L2_CAP_TUNER | V4L2_CAP_AUDIO
                      | V4L2_CAP_READWRITE | V4L2_CAP_STREAMING,
    .reserved	    = {0,0,0,0}
};

static struct v4l2_tuner pvr_v4l2_tuners[]= {
    {
        .index      = 0,
        .name       = "TV Tuner",
        .type		= V4L2_TUNER_ANALOG_TV,
        .capability	= V4L2_TUNER_CAP_NORM | V4L2_TUNER_CAP_STEREO |
                      V4L2_TUNER_CAP_LANG1 | V4L2_TUNER_CAP_LANG2,
        .rangelow   = 0,
        .rangehigh  = 0,
        .rxsubchans	= V4L2_TUNER_SUB_STEREO,
        .audmode	= V4L2_TUNER_MODE_STEREO,
        .signal		= 0,
        .afc		= 0,
        .reserved	= {0,0,0,0}
    }
/*    {
        .index	= 1,
        .name   = "Radio Tuner",
        .type		= V4L2_TUNER_RADIO,
        .capability	= (V4L2_TUNER_CAP_STEREO),
        .rangelow   = 0,
        .rangehigh  = 0,
        .rxsubchans	= 0,
        .audmode	= V4L2_TUNER_MODE_STEREO,
        .signal		= 0,
        .afc		= 0,
        .reserved	= {0,0,0,0}
    }*/
};

struct v4l2_standard pvr_standards[] = {
    {
        .index          = 0,
        .id             = V4L2_STD_PAL_BG,
        .name           = "PAL(B/G)",
        .frameperiod    =
            {
                .numerator	= 1,
                .denominator= 25
            },
        .framelines     = 625,
        .reserved       = {0,0,0,0}
	},{
        .index          = 1,
        .id             = V4L2_STD_PAL_I,
        .name           = "PAL(I)",
        .frameperiod    =
            {
                .numerator	= 1,
                .denominator= 25
            },
        .framelines     = 625,
        .reserved       = {0,0,0,0}
	},{
        .index          = 2,
        .id             = V4L2_STD_PAL_DK,
        .name           = "PAL(D/K)",
        .frameperiod    =
            {
                .numerator	= 1,
                .denominator= 25
            },
        .framelines     = 625,
        .reserved       = {0,0,0,0}
	},{
        .index          = 3,
        .id             = V4L2_STD_SECAM,
        .name           = "SECAM(L/L')",
        .frameperiod    =
            {
                .numerator	= 1,
                .denominator= 25
            },
        .framelines     = 625,
        .reserved       = {0,0,0,0}
	},
    {
	    .index          = 4,
	    .id             = V4L2_STD_NTSC,
	    .name           = "NTSC(M)",
	    .frameperiod	=
            {
                .numerator =  1001,
                .denominator= 30000
            },
	    .framelines	= 525,
	    .reserved	= {0,0,0,0}
	}
};

u32 pvr_tbl_standard[] = {
    0x00d47009, /* PVR_STD_PAL_BG */
    0x00d4700a, /* PVR_STD_PAL_I */
    0x00d4700b, /* PVR_STD_PAL_DK */
    0x00c4100b, /* PVR_STD_SECAM */
    0x00163044  /* PVR_STD_NTSC */
};

static struct v4l2_input pvr_v4l2_inputs[]= {
    {
        .index      = 0,
        .name		= "TV",
        .type		= V4L2_INPUT_TYPE_TUNER,
        .audioset	= 0,
        .tuner	    = 0,
        .std        = 0, /* filled by each driver instance depending on eeprom data*/
        .status	    = 0,
    },{
        .index	    = 1,
        .name		= "Composite",
        .type		= V4L2_INPUT_TYPE_CAMERA,
        .audioset	= 1,
        .tuner	    = 0,
        .std	    = 0,
        .status	    = 0,
    },{
        .index	    = 2,
        .name		= "S-Video",
        .type		= V4L2_INPUT_TYPE_CAMERA,
        .audioset	= 1,
        .tuner	    = 0,
        .std	    = 0,
        .status	    = 0,
    }
/*        .index	= 2,
        .name		= "Radio Tuner",
        .type		= V4L2_INPUT_TYPE_TUNER,
        .audioset	= 2,
        .tuner	    = 1,
        .std        = 0
        .status	    = 0,
    }
*/
};

struct v4l2_audio pvr_audio_inputs[] = {
        {
         .index         = 0,
         .name          = "TV Audio", 
         .capability    = 0,
         .mode          = 0,
         .reserved      = {0,0} 
        }, {
         .index         = 1,
         .name          = "RCA Audio",
         .capability    = 0,
         .mode          = 0,
         .reserved      = {0,0} 
        }
};

struct v4l2_fmtdesc pvr_fmtdesc [] = {
    {
        .index          = 0,
        .type           = V4L2_BUF_TYPE_VIDEO_CAPTURE,
        .flags          = V4L2_FMT_FLAG_COMPRESSED,
        .description    = "MPEG1/2",
        .pixelformat    = 0, // V4L2_PIX_FMT_MPEG,
        .reserved       = { 0, 0, 0, 0 }
    }
};

struct v4l2_format pvr_format [] = {
    {
        .type	= V4L2_BUF_TYPE_VIDEO_CAPTURE,
        .fmt	= {
            .pix	= {
                .width 	        = 720,
                .height 	    = 576,
                .pixelformat    = 0, //V4L2_PIX_FMT_MPEG,
                .field          = V4L2_FIELD_INTERLACED,
                .bytesperline   = 0,  // doesn't make sense here  
                .sizeimage	    = (32*1024), //FIXME : Don't know what to put here...
                .colorspace     = 0, // doesn't make sense here
                .priv           = 0
            }
        }
    },{
        .type	= V4L2_BUF_TYPE_VBI_CAPTURE,
        .fmt	= {
            .vbi	= {
                .sampling_rate = 27000000,
                .offset = 248,
                .samples_per_line = 1443,
                .sample_format = V4L2_PIX_FMT_GREY,
                .start = { 0, 0 },
                .count = { 0, 0 },
                .flags = 0,
                .reserved = { 0, 0 }
            }
        }
    }
};

struct v4l2_queryctrl pvr_regular_ctl [] = 
{
    [V4L2_CID_BRIGHTNESS - V4L2_CID_BASE] = {
        .id = V4L2_CID_BRIGHTNESS,
        .type = V4L2_CTRL_TYPE_INTEGER,
        .name = "Brightness",
        .minimum = 0,
        .maximum = 255,
        .step = 1,
        .default_value = 128,
        .flags = 0,
        .reserved = { 0, 0 }
    },
    [V4L2_CID_CONTRAST - V4L2_CID_BASE] = {
        .id = V4L2_CID_CONTRAST,
        .type = V4L2_CTRL_TYPE_INTEGER,
        .name = "Contrast",
        .minimum = 0,
        .maximum = 127,
        .step = 1,
        .default_value = 68,
        .flags = 0,
        .reserved = { 0, 0 }
        
    },
    [V4L2_CID_SATURATION - V4L2_CID_BASE] = {
        .id = V4L2_CID_SATURATION,
        .type = V4L2_CTRL_TYPE_INTEGER,
        .name = "Saturation",
        .minimum = 0,
        .maximum = 127,
        .step = 1,
        .default_value = 64,
        .flags = 0,
        .reserved = { 0, 0 }
    },
    [V4L2_CID_HUE - V4L2_CID_BASE] = {
        .id = V4L2_CID_HUE,
        .type = V4L2_CTRL_TYPE_INTEGER,
        .name = "Hue",
        .minimum = -128,
        .maximum = 127,
        .step = 1,
        .default_value = 0,
        .flags = 0,
        .reserved = { 0, 0 }
    },
    [V4L2_CID_AUDIO_VOLUME - V4L2_CID_BASE] = {
        .id = V4L2_CID_AUDIO_VOLUME,
        .type = V4L2_CTRL_TYPE_INTEGER,
        .name = "Volume",
        .minimum = 0,
        .maximum = 0x7f,
        .step = 1,
        .default_value = 0x7e,
        .flags = 0,
        .reserved = { 0, 0 }
    },
    [V4L2_CID_AUDIO_BALANCE - V4L2_CID_BASE] = {
        .id = V4L2_CID_AUDIO_BALANCE,
        .type = V4L2_CTRL_TYPE_INTEGER,
        .name = "Balance",
        .minimum = -128,
        .maximum = 127,
        .step = 1,
        .default_value = 0,
        .flags = 0,
        .reserved = { 0, 0 }
    },
    [V4L2_CID_AUDIO_BASS - V4L2_CID_BASE] = {
        .id = V4L2_CID_AUDIO_BASS,
        .type = V4L2_CTRL_TYPE_INTEGER,
        .name = "Bass",
        .minimum = -96,
        .maximum = 127,
        .step = 0,
        .default_value = 0,
        .flags = 0,
        .reserved = { 0, 0 }
    },
    [V4L2_CID_AUDIO_TREBLE - V4L2_CID_BASE] = {
        .id = V4L2_CID_AUDIO_TREBLE,
        .type = V4L2_CTRL_TYPE_INTEGER,
        .name = "Treble",
        .minimum = -96,
        .maximum = 120,
        .step = 0,
        .default_value = 0,
        .flags = 0,
        .reserved = { 0, 0 }
    },
    [V4L2_CID_AUDIO_MUTE - V4L2_CID_BASE] = {
        .id = V4L2_CID_AUDIO_MUTE,
        .type = V4L2_CTRL_TYPE_INTEGER,
        .name = "Mute",
        .minimum = 0,
        .maximum = 1,
        .step = 1,
        .default_value = 0,
        .flags = 0,
        .reserved = { 0, 0 }
    },
};

/* Now our private controls. */

struct v4l2_queryctrl pvr_private_ctl [] =
{
   [V4L2_CID_PVR_SRATE - V4L2_CID_PRIVATE_BASE] = {
        .id = V4L2_CID_PVR_SRATE,
        .type = V4L2_CTRL_TYPE_MENU,
        .name = "Sample rate",
        .minimum = 0,
        .maximum = 1,
        .step = 1,
        .default_value = 0,
        .flags = 0,
        .reserved = { 0, 0 }
   },
   [V4L2_CID_PVR_AUDIOBITRATE - V4L2_CID_PRIVATE_BASE] = {
        .id = V4L2_CID_PVR_AUDIOBITRATE,
        .type = V4L2_CTRL_TYPE_MENU,
        .name = "Audio Bitrate",
        .minimum = 0,
        .maximum = 4,   /* only a few bitrates seem to be supported */
        .step = 1,
        .default_value = 3,
        .flags = 0,
        .reserved = { 0, 0 }
   },
   [V4L2_CID_PVR_AUDIOCRC - V4L2_CID_PRIVATE_BASE] = {
        .id = V4L2_CID_PVR_AUDIOCRC,
        .type = V4L2_CTRL_TYPE_MENU,
        .name = "Audio CRC",
        .minimum = 0,
        .maximum = 1,
        .step = 1,
        .default_value = 0,
        .flags = 0,
        .reserved = { 0, 0 }
   },
   [V4L2_CID_PVR_AUDIOEMPHASIS - V4L2_CID_PRIVATE_BASE] = {
        .id = V4L2_CID_PVR_AUDIOEMPHASIS,
        .type = V4L2_CTRL_TYPE_MENU,
        .name = "Audio Emphasis",
        .minimum = 0,
        .maximum = 2,
        .step = 1,
        .default_value = 0,
        .flags = 0,
        .reserved = { 0, 0 }
   },
   [V4L2_CID_PVR_VBR - V4L2_CID_PRIVATE_BASE] = {
        .id = V4L2_CID_PVR_VBR,
        .type = V4L2_CTRL_TYPE_MENU,
        .name = "Variable video bitrate",
        .minimum = 0,
        .maximum = 1,
        .step = 1,
        .default_value = 1,
        .flags = 0,
        .reserved = { 0, 0 }
   },
   [V4L2_CID_PVR_VIDEOBITRATE - V4L2_CID_PRIVATE_BASE] = {
        .id = V4L2_CID_PVR_VIDEOBITRATE,
        .type = V4L2_CTRL_TYPE_INTEGER,
        .name = "Average video bitrate",
        .minimum = 1,
        .maximum = 20000000,
        .step = 1,
        .default_value = 6000000,
        .flags = 0,
        .reserved = { 0, 0 }
   },
   [V4L2_CID_PVR_VIDEOPEAK - V4L2_CID_PRIVATE_BASE] = {
        .id = V4L2_CID_PVR_VIDEOPEAK,
        .type = V4L2_CTRL_TYPE_INTEGER,
        .name = "Peak video bitrate",
        .minimum = 1,
        .maximum = 20000000,
        .step = 400,
        .default_value = 6000000,
        .flags = 0,
        .reserved = { 0, 0 }
   },
};

struct v4l2_querymenu pvr_ctl_query_srate[] = {
	/* ID, Index, Name, Reserved */
	{V4L2_CID_PVR_SRATE, 0, "48kHz",   0 },
	{V4L2_CID_PVR_SRATE, 1, "44.1kHz", 0 },
};

#define PVR_SRATE_48 0
#define PVR_SRATE_44 1

u32 pvr_tbl_srate[] = { 0x01, 0x00,	0x02 };

struct v4l2_querymenu pvr_ctl_query_audiobitrate[] = {
	/* ID, Index, Name, Reserved */
    {V4L2_CID_PVR_AUDIOBITRATE,  0, "384k", 0},
    {V4L2_CID_PVR_AUDIOBITRATE,  1, "320k", 0},
    {V4L2_CID_PVR_AUDIOBITRATE,  2, "256k", 0},
	{V4L2_CID_PVR_AUDIOBITRATE,  3, "224k", 0},
	{V4L2_CID_PVR_AUDIOBITRATE,  4, "192k", 0},
	{V4L2_CID_PVR_AUDIOBITRATE,  5, "160k", 0},
	{V4L2_CID_PVR_AUDIOBITRATE,  6, "128k", 0},
	{V4L2_CID_PVR_AUDIOBITRATE,  7, "112k", 0},
	{V4L2_CID_PVR_AUDIOBITRATE,  8, "96k",  0},
	{V4L2_CID_PVR_AUDIOBITRATE,  9, "80k",  0},
	{V4L2_CID_PVR_AUDIOBITRATE, 10, "64k",  0},
	{V4L2_CID_PVR_AUDIOBITRATE, 11, "56k",   0},
	{V4L2_CID_PVR_AUDIOBITRATE, 12, "48k",   0},
	{V4L2_CID_PVR_AUDIOBITRATE, 13, "32k",   0},
	{V4L2_CID_PVR_AUDIOBITRATE, 14, "Free fmt",  0}
};

u32 pvr_tbl_audiobitrate[] = {
	/* setting */
	0xe << 4, /* 384k */
	0xd << 4, /* 320k */
	0xc << 4, /* 256k */
	0xb << 4, /* 224k */
	0xa << 4, /* 192k */
	0x9 << 4, /* 160k */
	0x8 << 4, /* 128k */
	0x7 << 4, /* 112k */
    0x6 << 4, /* 96k  */
	0x5 << 4, /* 80k  */
	0x4 << 4, /* 64k  */
	0x3 << 4, /* 56k   */
	0x2 << 4, /* 48k   */
	0x1 << 4, /* 32k   */
    0x0 << 4  /*  fmt  */
};

struct v4l2_querymenu pvr_ctl_query_crc[] = {
	/* ID, Index, Name, Reserved */
	{V4L2_CID_PVR_AUDIOCRC, 0, "On",  0},
	{V4L2_CID_PVR_AUDIOCRC, 1, "Off", 0},
};

u32 pvr_tbl_crc [] = { 0x1 << 14, 0x0 << 14};

struct v4l2_querymenu pvr_ctl_query_emphasis[] = {
	/* ID, Index, Name, Reserved */
	{V4L2_CID_PVR_AUDIOEMPHASIS, 0, "None",  0},
	{V4L2_CID_PVR_AUDIOEMPHASIS, 1, "50/15uS", 0},
    {V4L2_CID_PVR_AUDIOEMPHASIS, 2, "CCITT J.17", 0},
};

u32 pvr_tbl_emphasis [] = { 0x0 << 12, 0x1 << 12, 0x3 << 12 };

struct v4l2_querymenu pvr_ctl_query_vbr[] = {
	/* ID, Index, Name, Reserved */
	{V4L2_CID_PVR_VBR, 0, "On",  0},
	{V4L2_CID_PVR_VBR, 1, "Off", 0},
};

u32 pvr_tbl_vbr [] = { 0x01, 0x00 };

struct v4l2_querymenu * pvr_private_menu[] = {
    [V4L2_CID_PVR_SRATE - V4L2_CID_PRIVATE_BASE] = pvr_ctl_query_srate,
    [V4L2_CID_PVR_AUDIOBITRATE - V4L2_CID_PRIVATE_BASE] = pvr_ctl_query_audiobitrate,
    [V4L2_CID_PVR_AUDIOCRC - V4L2_CID_PRIVATE_BASE] = pvr_ctl_query_crc,
    [V4L2_CID_PVR_AUDIOEMPHASIS - V4L2_CID_PRIVATE_BASE] = pvr_ctl_query_emphasis,
    [V4L2_CID_PVR_VBR - V4L2_CID_PRIVATE_BASE] = pvr_ctl_query_vbr
};

/*
 * pvr_ioctl()
 *
 * This is part of Video 4 Linux API. The procedure handles ioctl() calls. 
 *
 */
int pvr_v4l_do_ioctl(struct inode *inode, struct file *file,
				 unsigned int cmd, void *arg)
{
	struct pvr_fh *fh = file->private_data;
    struct pvr_stream *stream = fh->stream;
    struct pvr *pvr = stream->pvr;
    int ret = -EINVAL, dirty = 0;

    pvr_debug("pvr_v4l_do_ioctl");

    if ((pvr == NULL) || (pvr->udev == NULL) || pvr->disconnecting)
		return -EFAULT;

    /* check priority */
	switch (cmd) {
	case VIDIOC_S_CTRL:
	case VIDIOC_S_STD:
	case VIDIOC_S_INPUT:
	case VIDIOC_S_TUNER:
	case VIDIOC_S_FREQUENCY:
		ret = v4l2_prio_check(&pvr->prio, &fh->prio);
		if (ret)
			return ret;
    }

    switch (cmd) {
        case VIDIOC_QUERYCAP:
		{
			struct v4l2_capability *cap = arg;
            
            memcpy(cap, &pvr_capability, sizeof(struct v4l2_capability));

            ret = 0;
            break;
		}

        case VIDIOC_G_PRIORITY:
        {
            enum v4l2_priority *p = arg;

            *p = v4l2_prio_max(&pvr->prio);
            ret = 0;
            break;
        }

        case VIDIOC_S_PRIORITY:
        {
            enum v4l2_priority *prio = arg;

            ret = v4l2_prio_change(&pvr->prio, &fh->prio, *prio);
            break;
        }

        case VIDIOC_ENUMSTD:
        {

    	    struct v4l2_standard *vs = (struct v4l2_standard *)arg;

		    if ((vs->index < 0) || (vs->index >= PVR_V4L2_NSTD))
                break;
		
            memcpy(vs, &pvr_standards[vs->index], sizeof(struct v4l2_standard));

		    ret = 0;
            break;
        }
		
        case VIDIOC_G_STD:
        {
    	    v4l2_std_id *vs = (v4l2_std_id *)arg;

            PVR_BUG_ON(pvr->video.standard < 0 || pvr->video.standard >= PVR_V4L2_NSTD);

		    *vs = pvr_standards[pvr->video.standard].id;
		    ret = 0;
            break;
        }

        case VIDIOC_S_STD:
        {
    	    v4l2_std_id *vs = (v4l2_std_id *)arg;

            if (*vs & V4L2_STD_PAL_BG){
                dirty = (pvr->video.standard != PVR_STD_PAL_BG);
                pvr->video.standard = PVR_STD_PAL_BG;
                ret = 0;
            } else if (*vs & V4L2_STD_PAL_I){
                dirty = (pvr->video.standard != PVR_STD_PAL_I);
                pvr->video.standard = PVR_STD_PAL_I;
                ret = 0;
            } else if (*vs & V4L2_STD_PAL_DK){
                dirty = (pvr->video.standard != PVR_STD_PAL_DK);
                pvr->video.standard = PVR_STD_PAL_DK;
                ret = 0;                
            } else if (*vs & V4L2_STD_SECAM){
                dirty = (pvr->video.standard != PVR_STD_SECAM);
                pvr->video.standard = PVR_STD_SECAM;
                ret = 0;                
            } else if (*vs & V4L2_STD_NTSC){
                dirty = (pvr->video.standard != PVR_STD_NTSC);
                pvr->video.standard = PVR_STD_NTSC;
                ret = 0;                
            }

            break;
        }
        
        case VIDIOC_ENUMINPUT:
        {
            struct v4l2_input *vi = (struct v4l2_input *)arg;
            
            pvr_debug("VIDIOC_ENUM_INPUT");

            if ((vi->index >= PVR_V4L2_NINPUT) || (vi->index < 0))
                break;
            
            memcpy(vi, &pvr_v4l2_inputs[vi->index], sizeof(struct v4l2_input));
            
            ret = 0;
            break;            
        }

        case VIDIOC_G_INPUT:
        {
            struct v4l2_input *vi = (struct v4l2_input *)arg;

            pvr_debug("VIDIOC_G_INPUT");

            if (pvr->audio.input != pvr->video.input)
                err("audio and video sources don't match");
            vi->index = pvr->video.input;
            ret = 0;
            break;
        }

        case VIDIOC_S_INPUT:
        {
            struct v4l2_input *vi = (struct v4l2_input *)arg;

            pvr_debug("VIDIOC_S_INPUT");

            if ((vi->index >= PVR_V4L2_NINPUT) || (vi->index < 0))
                break;
            pvr->video.input = vi->index;
            pvr->audio.input = vi->index;
            ret = pvr_decoder_set_input(pvr);
            ret |= pvr_configure_audio(pvr);

            break;
        }

        case VIDIOC_ENUMAUDIO:
        {
            struct v4l2_audio *va = (struct v4l2_audio *)arg;
            
            if ((va->index >= PVR_V4L2_AUDIO_NINPUT) || (va->index < 0))
                break;
            
            memcpy(va, &pvr_audio_inputs[va->index],
                   sizeof(struct v4l2_audio));
            
            ret = 0;
            break;            
        }
        
        case VIDIOC_G_AUDIO:
        {
        struct v4l2_audio *va = (struct v4l2_audio *)arg;

            memcpy(va, &pvr_audio_inputs[pvr->audio.input],
                   sizeof(struct v4l2_audio));
            ret = 0;
            break;
        }
        
        case VIDIOC_S_AUDIO:
        {
           // struct v4l2_audio *va = (struct v4l2_audio *)arg;

           break;
        }
        case VIDIOC_G_TUNER:
        {
            struct v4l2_tuner *vt = (struct v4l2_tuner *)arg;

            pvr_debug("VIDIOC_G_TUNER");

            if (vt->index !=0)
                break;

            memcpy(vt, &pvr_v4l2_tuners[vt->index], sizeof(struct v4l2_tuner));

            if (pvr_audio_update_status(pvr) == 0){
                if (pvr->audio.status_stereo)
                    vt->rxsubchans = V4L2_TUNER_SUB_STEREO;
                else    
                    vt->rxsubchans = V4L2_TUNER_SUB_MONO;
                if (pvr->audio.status_bilingual)
                    vt->rxsubchans |= V4L2_TUNER_SUB_LANG1 |
                                      V4L2_TUNER_SUB_LANG2;
            }

            if (pvr->audio.input == PVR_INPUT_TV)
                vt->audmode = pvr->audio.mode;
            else
                vt->audmode = V4L2_TUNER_MODE_STEREO;

            vt->signal = pvr_decoder_is_tuned(pvr) ? 65535 : 0;

            ret = 0;
            break;
        }

        case VIDIOC_S_TUNER:
        {
            struct v4l2_tuner *vt=(struct v4l2_tuner *)arg;

            pvr_debug("VIDIOC_S_TUNER");

            if (vt->index != 0)
                break;

            pvr->audio.mode = vt->audmode;
            ret = pvr_audio_set_stereo(pvr);
        }

        case VIDIOC_G_FREQUENCY:
        {
            /* FIXME : suppport radio tuner */
            struct v4l2_frequency *vf = (struct v4l2_frequency *)arg;

            pvr_debug("VIDIOC_G_FREQUENCY");

            if (vf->tuner != 0)
                break;

            if (stream->v4l_type == VFL_TYPE_RADIO)
                vf->frequency = pvr->radio_frequency;
            else
                vf->frequency = pvr->tv_frequency;

            ret = 0;
            break;
        }

        case VIDIOC_S_FREQUENCY:
        {
            /* FIXME : suppport radio tuner */
            struct v4l2_frequency *vf = (struct v4l2_frequency *)arg;

            pvr_debug("VIDIOC_S_FREQUENCY");

            if (vf->tuner != 0)
                break;
            if (stream->v4l_type == VFL_TYPE_RADIO) {
                pvr->radio_frequency = vf->frequency;
                ret = pvr_set_freq_radio(pvr);
            } else {
                pvr->tv_frequency = vf->frequency;
                ret = pvr_set_freq_tv(pvr);
                /*  FIXME : needed ? If not we shall at least sleep a bit. */
                ret |= pvr_configure_audio(pvr);
            }

            break;
        }

        case VIDIOC_ENUM_FMT:
        {
            struct v4l2_fmtdesc *fd = (struct v4l2_fmtdesc *)arg;


            pvr_debug("VIDIOC_ENUM_FMT");
            /* Only one format is supported : mpeg.*/
            if (fd->index != 0)
                break;

            memcpy(fd, pvr_fmtdesc, sizeof(struct v4l2_fmtdesc));
            ret = 0;
            break;
        }

        case VIDIOC_G_FMT:
        {
            struct v4l2_format *vf = (struct v4l2_format *)arg;
            int is_ntsc = (pvr->video.standard == PVR_STD_NTSC);
            int height = is_ntsc ? 480 : 576;

            pvr_debug("VIDIOC_G_FMT");

            ret = 0;
            switch(vf->type) {
                case V4L2_BUF_TYPE_VIDEO_CAPTURE:
                    memcpy(vf, &pvr_format[PVR_FORMAT_PIX],
                           sizeof(struct v4l2_format));
                    vf->fmt.pix.width = pvr->video.width;
                    vf->fmt.pix.height = pvr->video.height_full ? height : height / 2;
                    break;
                case V4L2_BUF_TYPE_VBI_CAPTURE:
                    memcpy(vf, &pvr_format[PVR_FORMAT_VBI],
                           sizeof(struct v4l2_format));
                    vf->fmt.vbi.start[0] = is_ntsc ? 10 : 6;
                    vf->fmt.vbi.start[1] = is_ntsc ? 273 : 318;
                    vf->fmt.vbi.count[0] = is_ntsc ? 12 : 18;
                    vf->fmt.vbi.count[1] = is_ntsc ? 12 : 18;
                    break;
                default:
                    ret = -EINVAL;
                    break;
            }
            break;
        }
        
        case VIDIOC_TRY_FMT:
        case VIDIOC_S_FMT:
        {
            struct v4l2_format *vf = (struct v4l2_format *)arg;
            int is_ntsc = (pvr->video.standard == PVR_STD_NTSC);

            pvr_debug("VIDIOC_{S,TRY}_FMT");

            ret = 0;
            if (vf->type == V4L2_BUF_TYPE_VBI_CAPTURE){
                memcpy(vf, &pvr_format[PVR_FORMAT_VBI],
                      sizeof(struct v4l2_format));
                vf->fmt.vbi.start[0] = is_ntsc ? 10 : 6;
                vf->fmt.vbi.start[1] = is_ntsc ? 273 : 318;
                vf->fmt.vbi.count[0] = is_ntsc ? 12 : 18;
                vf->fmt.vbi.count[1] = is_ntsc ? 12 : 18;
            } else if (vf->type == V4L2_BUF_TYPE_VIDEO_CAPTURE){
                int hf = is_ntsc ? 480 : 576;
                int hh = (int) (hf / 2);
                int h = vf->fmt.pix.height;
                int w = vf->fmt.pix.width;

                memcpy(vf, &pvr_format[PVR_FORMAT_PIX],
                       sizeof(struct v4l2_format));
                if (w > 720)
                    vf->fmt.pix.width = 720;
                vf->fmt.pix.width &= 0xff0;
                vf->fmt.pix.height = (h > hh) ? hf : hh; 

                if (cmd == VIDIOC_S_FMT){
                    down(&pvr->sem);                
                    pvr->video.width = vf->fmt.pix.width;
                    pvr->video.height_full = (vf->fmt.pix.height == hf);
                    up(&pvr->sem);  
                    dirty = 1;
                }
            } else
                ret = -EINVAL;
            break;
        }

        case VIDIOC_QBUF:
        {
            struct v4l2_buffer *vb = (struct v4l2_buffer *) arg;

            ret = pvr_stream_qbuf(stream, vb);
            break;
        }
        case VIDIOC_QUERYBUF:
        {
            struct v4l2_buffer *vb = (struct v4l2_buffer *) arg;

            ret = pvr_stream_querybuf(stream, vb);
            break;
        }
        case VIDIOC_DQBUF:
        {
            struct v4l2_buffer *vb = (struct v4l2_buffer *) arg;
 
            ret = pvr_stream_dqbuf(stream, vb, file->f_flags & O_NONBLOCK);
            break;
        }
        case VIDIOC_REQBUFS:
        {
            struct v4l2_requestbuffers *vr = (struct v4l2_requestbuffers *) arg;

            ret = pvr_stream_reqbufs(stream, fh, vr);
            break;
        }

        case VIDIOC_STREAMON:
        {
            ret = pvr_stream_streamon(stream, fh);
            break;
        }

        case VIDIOC_STREAMOFF:
        {
            ret = pvr_stream_streamoff(stream, fh);
            break;
        }
        
        case VIDIOC_QUERYCTRL:
        {
            struct v4l2_queryctrl *vc = (struct v4l2_queryctrl *)arg;

            ret = 0;
            switch(vc->id){
                case V4L2_CID_BRIGHTNESS:
                case V4L2_CID_SATURATION:
                case V4L2_CID_CONTRAST:
                case V4L2_CID_HUE:
                case V4L2_CID_AUDIO_VOLUME:
                case V4L2_CID_AUDIO_BALANCE:
                case V4L2_CID_AUDIO_BASS:
                case V4L2_CID_AUDIO_TREBLE:
                case V4L2_CID_AUDIO_MUTE:
                    memcpy(vc ,&pvr_regular_ctl[vc->id - V4L2_CID_BASE],
                           sizeof(struct v4l2_queryctrl));
                break;
                default:
                    ret = -EINVAL;
           }
            
            if ((vc->id >= V4L2_CID_PRIVATE_BASE) &&
                (vc->id <= V4L2_CID_PVR_MAX)){
                int i = vc->id - V4L2_CID_PRIVATE_BASE;
                memcpy(vc, &pvr_private_ctl[i], sizeof(struct v4l2_queryctrl));
                ret = 0;
            }
            
            break;
        }

        case VIDIOC_QUERYMENU:
        {
            struct v4l2_querymenu *vm = (struct v4l2_querymenu *)arg;

            if ((vm->id >= V4L2_CID_PRIVATE_BASE) &&
                (vm->id <= V4L2_CID_PVR_MAX)){
                int i = vm->id - V4L2_CID_PRIVATE_BASE;
                if ((vm->index >= 0) &&
                    (vm->index <= pvr_private_ctl[i].maximum)){
                    memcpy(vm, &pvr_private_menu[i][vm->index],
                           sizeof(struct v4l2_querymenu));
                    ret = 0;
                }
            }

            
            break;
        }

        case VIDIOC_G_CTRL:
        {
            struct v4l2_control *vc = (struct v4l2_control *)arg;

            ret = 0;
            switch(vc->id){
                case V4L2_CID_PVR_SRATE:
                    vc->value = pvr->audio.srate;
                break;
                case V4L2_CID_PVR_AUDIOBITRATE:
                    vc->value = pvr->audio.bitrate;
                break;
                case V4L2_CID_PVR_AUDIOCRC:
                    vc->value = pvr->audio.crc;
                break;
                case V4L2_CID_PVR_AUDIOEMPHASIS:
                    vc->value = pvr->audio.emphasis;
                break;
                case V4L2_CID_PVR_VBR:
                    vc->value = pvr->video.vbr;
                break;
                case V4L2_CID_PVR_VIDEOBITRATE:
                    vc->value = pvr->video.bitrate;
                break;
                case V4L2_CID_PVR_VIDEOPEAK:
                    vc->value = pvr->video.peakbitrate;
                break;
                case V4L2_CID_BRIGHTNESS:
                    vc->value = pvr->video.brightness;
                break;
                case V4L2_CID_CONTRAST:
                    vc->value = pvr->video.contrast;
                break;
                case V4L2_CID_SATURATION:
                    vc->value = pvr->video.saturation;
                break;
                case V4L2_CID_HUE:
                    vc->value = pvr->video.hue;
                break;
                case V4L2_CID_AUDIO_VOLUME:
                    vc->value = pvr->audio.volume;
                break;
                case V4L2_CID_AUDIO_BALANCE:
                    vc->value = pvr->audio.balance;
                break;
                case V4L2_CID_AUDIO_BASS:
                    vc->value = pvr->audio.bass;
                break;
                case V4L2_CID_AUDIO_TREBLE:
                    vc->value = pvr->audio.treble;
                break;
                case V4L2_CID_AUDIO_MUTE:
                    vc->value = pvr->audio.muted;
                break;
                default:
                    ret = -EINVAL;
           }
            break;
        }

        case VIDIOC_S_CTRL:
        {
            struct v4l2_control *vc = (struct v4l2_control *)arg;

#define PVR_CHECK_INVAL(min, max, v)   if ((v) < (min) || (v) > (max)) { \
                                           ret = -ERANGE; \
                                           break; \
                                       }
            ret = 0;
            switch(vc->id){
                case V4L2_CID_PVR_SRATE:
                    dirty = (pvr->audio.srate != vc->value);
                    pvr->audio.srate = vc->value;
                break;
                case V4L2_CID_PVR_AUDIOBITRATE:
                    dirty = (pvr->audio.bitrate != vc->value);
                    pvr->audio.bitrate = vc->value;
                break;
                case V4L2_CID_PVR_AUDIOCRC:
                    dirty = (pvr->audio.crc != vc->value);
                    pvr->audio.crc = vc->value;
                break;
                case V4L2_CID_PVR_AUDIOEMPHASIS:
                    dirty = (pvr->audio.emphasis != vc->value);
                    pvr->audio.emphasis = vc->value;
                break;
                case V4L2_CID_PVR_VBR:
                    dirty = (pvr->video.vbr != vc->value);
                    pvr->video.vbr = vc->value;
                break;
                case V4L2_CID_PVR_VIDEOBITRATE:
                    dirty = (pvr->video.bitrate != vc->value);
                    pvr->video.bitrate = vc->value;
                break;
                case V4L2_CID_PVR_VIDEOPEAK:
                    dirty = (pvr->video.peakbitrate != vc->value);
                    pvr->video.peakbitrate = vc->value;
                break;
                case V4L2_CID_BRIGHTNESS:
                    PVR_CHECK_INVAL(0, 255, vc->value);
                    pvr->video.brightness = vc->value;
                    ret = pvr_decoder_set_bcsh(pvr);
                break;
                case V4L2_CID_CONTRAST:
                    PVR_CHECK_INVAL(0, 127, vc->value);
                    pvr->video.contrast = vc->value;
                    ret = pvr_decoder_set_bcsh(pvr);
                break;
                case V4L2_CID_SATURATION:
                    PVR_CHECK_INVAL(0, 127, vc->value);
                    pvr->video.saturation = vc->value;
                    ret = pvr_decoder_set_bcsh(pvr);
                break;
                case V4L2_CID_HUE:
                    PVR_CHECK_INVAL(-128, 127, vc->value);
                    pvr->video.hue = vc->value;
                    ret = pvr_decoder_set_bcsh(pvr);
                break;
                case V4L2_CID_AUDIO_VOLUME:
                    PVR_CHECK_INVAL(0, 65535, vc->value);
                    pvr->audio.volume = vc->value;
                    ret = pvr_audio_setvolume(pvr);
                break;
                case V4L2_CID_AUDIO_BALANCE:
                    PVR_CHECK_INVAL(-65536, 65535, vc->value);
                    pvr->audio.balance = vc->value;
                    ret = pvr_audio_setvolume(pvr);
                break;
                case V4L2_CID_AUDIO_MUTE:
                    pvr->audio.muted = vc->value;
                    ret = pvr_audio_setvolume(pvr);
                break;
                default:
                    ret = -EINVAL;
            }
            break;
        }

        /* v4l1 compatibility */
        case VIDIOCGMBUF:
        {
            struct video_mbuf *vmb = (struct video_mbuf *) arg;

            ret = pvr_stream_cgmbuf(stream, fh ,vmb);
            break;
        }

        default :
            ret = v4l_compat_translate_ioctl(inode, file, cmd, arg, pvr_v4l_do_ioctl);
    }

    if (dirty)
        pvr_reconfigure(pvr);

    return ret;
}
